from collections import OrderedDict

SINGLE_CHOICE_TYPE = 'single_choice'
MULTIPLE_CHOICE_TYPE = 'multiple_choice'
MATRIX_CHOICE_TYPE = 'matrix'
TEXT_TYPE = 'text'
NUMERIC_TYPE = 'numeric'
DATE_TYPE = 'date'

class StandardParserException(Exception):
    pass

class StandardSurvey:

    def __init__(self, name):
        self.name = name
        self.questions = OrderedDict()
        self.comment = None

    def add_question(self, question):
        n = question.data_name
        if n in self.questions:
            raise StandardParserException("Question with data_name '%s' is already defined" % (n))
        self.questions[n] = question

    def __repr__(self):
        qq = []
        for q in self.questions:
            qq.append(q.__repr__())
        return '<Survey %s %s>' % (self.name, ",".join(qq), )

class StandardQuestion:
    
    def __init__(self, data_name, type, title):
        self.data_name = data_name
        self.data_type = None
        self.type = type
        self.mandatory = False
        self.title = title
        self.comment = None
        self.data_type = None
        self.responses = StandardResponseList()
        self.rows = StandardMatrixDimensionList()
        self.columns = StandardMatrixDimensionList()
        self.order = 0
        self.active = True
        self.plateforms = []
        self.target = None
    
    def add_response(self, response):
        self.responses[response.key] = response
    
    def set_rows(self, rows):
        self.rows = rows
    
    def set_columns(self, columns):
        self.columns = columns

class StandardResponseList(OrderedDict):

    def find_by(self, name, value):
        for k, r in self.items():
            if hasattr(r, name):
                if getattr(r, name) == value:
                    return r
        return None

class StandardResponse:
      """
        key: local key in the spec
        text: label of the response
        order: order in the spec
        value: 
      """
      
      def __init__(self, key, text, order):
          self.key = key
          self.text = text
          self.order = order
          self.value = None
          self.comment = None
          self.active = True
          self.extra = None
          self.added_at = None
          self.removed_at = None
          self.plateforms = []


class StandardMatrixDimension:
    def __init__(self, text, key, value, order):
        """
            text: label
            key: key of the element in the definition (db id or json key dict)
            value: value to be used in the model
            order: order of the element in the definition
        """
        self.text = text
        self.key = key
        self.value = value
        self.order = order
        self.added_at = None
        self.removed_at = None

    def __repr__(self):
        return "[%s](%s)#%d" % (self.value, self.text, self.order)

class StandardMatrixDimensionList(OrderedDict):
    
    def find_by(self, name, value):
        for k, r in self.items():
            if hasattr(r, name):
                if getattr(r, name) == value:
                    return r
        return None




from .model import render_expression
from .parser import expression_parser

def readable_expression(expr, context):
    """
        Create a readable structure from an expression json object
        
        It turns an expression to a structure more easily readable once represented in yaml
    """
    ee = expression_parser(expr)
    return render_expression(ee, context)


from .parser import survey_parser, study_parser, Survey
from .dictionnary import ItemDictionnary
from ..readable import  as_readable
from typing import List, Optional
from ..context import Context

def readable_survey(survey, context:Context):
    """
        Transform a json based survey model into a 'readable' model (simpler and easier to read model) with a context
        
        context is a parameters set determining how to simplify the model (e.g. language..)
    """
    ss = survey_parser(survey)
    return as_readable(ss, context)

def render_to_dict(survey: Survey)-> Optional[ItemDictionnary]:
    """
        Render a Survey model into a dictionary based view (see influenzanet.dictionnary.ItemDictionnary)
    """
    item = survey.survey_definition
    return item.get_dictionnary()

def readable_study(study, context:Context):
    ss = study_parser(study)
    return as_readable(ss, context)

def survey_to_dictionnary(survey):
    """
        Transform a survey 
    """
    ss = survey_parser(survey)
    return render_to_dict(ss)


from .models import *

def json_parser_survey_standard(survey):
    """
        parse survey standard definition from json dictionnary
    """
    ss = StandardSurvey(survey['name'])
    if 'comment' in survey:
        ss.comment = json_parser_comment(survey['comment'])
    index = 1
    for qDef in survey['questions']:
        q = json_parser_question(qDef)
        q.order = index
        ss.add_question(q)
        index = index + 1
    return ss

def json_parser_comment(json):
    if isinstance(json, str):
        return [json]
    return json

def json_parser_description(json):
    if isinstance(json, str):
        return [json]
    return json

def import_attr(obj, data, keys):
    for key in keys:
        if key in data:
            setattr(obj, key, data[key])

def json_parser_question(json):
    """
        question definition from json dictionnary
    """
    data_name = json['data_name']

    q = StandardQuestion(data_name, json['type'], json['title'])
    
    if 'description' in json:
        q.description = json_parser_description(json['description'])
    
    if 'mandatory' in json:
        q.mandatory = json['mandatory']
  
    if 'comment' in json:
        q.comment = json_parser_comment(json['comment'])

    import_attr(q, json, ['active','data_type', 'format','rules', 'added_at', 'removed_at', 'platforms', 'target'])

    if 'possible_responses' in json:
        index = 1
        for key, rDef in json['possible_responses'].items():
            try:
                r = json_parser_response(key, rDef, index)
            except Exception as e:
                raise StandardParserException("Error in %s/%s : %s " % (data_name, key, e))
            q.add_response(r)
            index = index + 1

    if 'rows' in json:
        q.rows = json_parser_matrix_dim(json['rows'])

    if 'columns' in json:
        q.columns = json_parser_matrix_dim(json['columns'])

    return q

def json_parser_response(key, json, index):
    
    if not 'text' in json:
        raise StandardParserException("Missing 'text' field")
    
    r = StandardResponse(key, json['text'], index)
    
    import_attr(r, json, ['added_at', 'removed_at', 'platforms'])

    if 'value' in json:
        r.value = json['value']
    
    if 'description' in json:
        r.description = json_parser_description(json['description'])
    
    if 'comment' in json:
        r.comment = json_parser_comment(json['comment'])

    return r

def json_parser_matrix_dim(json):
    rr = StandardMatrixDimensionList()
    index = 1
    for key, d in json.items():
        #print(key, d)
        m = StandardMatrixDimension(d['text'], key, d['value'], index)
        rr[key] = m
        index = index + 1
    return rr


from __future__ import annotations

from ...context import Context
from . import KNOWN_EXPRESSIONS
from .types import Arg
from typing import Optional, List
class Expression:
    
    def __init__(self, name, params=Optional[List]):
        self.name = name
        if params is None:
            params = []
        self.params = params
        self._has_expression_param = False
        for p in params:
            if p.is_expression():
                self._has_expression_param = True
                break
    
    def is_expression(self):
        return True

    def has_expression_param(self):
        return self._has_expression_param

    def is_scalar(self):
        return False

    def param_name(self, index):
        if not self.name in KNOWN_EXPRESSIONS:
            return None
        expType = KNOWN_EXPRESSIONS[self.name]
        if expType.has_params():
            p = expType.get_param(index)
            if isinstance(p, Arg):
                return p.name
            return p

    def to_readable(self, ctx):
        return render_expression(self, ctx)

    def __repr__(self) -> str:
        return render_expression(self, Context()).__repr__()

class Scalar:

    def __init__(self, type, value):
        self.type = type
        self.value = value

    def is_expression(self):
        return False

    def is_scalar(self):
        return True

    def is_string(self):
        return self.type == 'str' or self.type == ''

    def is_numeric(self):
        return self.type == 'num'

    def __repr__(self):
        if self.type == "str":
            return '"' + self.value + '"'
        return str(self.value)
    
    def __str__(self):
        return str(self.value)


def with_default_names(args):
    """
    From a list of tuples (name, value) where name is the infered parameter name (None if not found)
    Transform the name of each by its index if is None
    """
    index = 0
    pp = []
    for a in args:
        if a[0] is None:
            a = ("%d" % (index), a[1])
        pp.append(a)
        index = index + 1
    return pp


def render_expression(ee, context):
    """
        Render an Expression object to a Yaml-ready readable data structure
    """
    if ee.is_expression():
        pp = []
        index = 0
        for p in ee.params:
            name = ee.param_name(index)
            pp.append( (name, render_expression(p, context)) )
            index = index + 1
        d = {}
        if ee.has_expression_param():
            pp = with_default_names(pp)
            pp = dict(pp)
            d[ee.name] = pp
        else:
            # All args are scalars
            ss = []
            for p in pp:
                if p[0] is not None:
                    s = "%s=%s" % (p[0], str(p[1]))
                else:
                    s = str(p[1])
                ss.append(s)
            pp = ', '.join(ss)
            d = "%s(%s)" % (ee.name, pp)
        return d
    if ee.is_scalar():
        return str(ee.value)
    return {'_unknown':'Unknown type'}  






from .influenzanet import  readable_survey, survey_to_dictionnary, survey_parser, readable_translatable, readable_study, read_survey_json
from .readable import as_readable
from .context import create_context, Context
from .html import survey_to_html
from .version import __version__




import unittest
from ....influenzanet.expression import Expression
from ....influenzanet.responses import RG_ROLES, RGROLES
from ....influenzanet.survey import ROLE_RESPONSE_GROUP, Survey, SurveyGroupItem, SurveyItemComponent, SurveyItemGroupComponent, SurveyItemResponseComponent, SurveySingleItem

def create_surveys(name, questions, selection=None):
    survey = Survey()
    
    items = []
    for quid, q in questions.items():
        
        comps = []
        if 'options' in q:
            rg = SurveyItemResponseComponent('rg', role=q['type'])
            rg_items = []
            for k,o in q['options'].items():
                rgi = SurveyItemComponent(k, 'option')
                rg_items.append(rgi)
            comps.append(rg_items)
                
        if len(comps) > 0:
            components=SurveyItemGroupComponent('root','root', comps, Expression('sequential'))
        else: 
            components = None
        validations = None
        item = SurveySingleItem(quid, components, validations, type=None)
        items.append(item)
    survey['surveyDefinition'] = SurveyGroupItem(name, items, selection=selection)
    
    return survey

class TestChecker(unittest.TestCase):

    def test_expression(self):
        survey = create_surveys('weekly', {
            'Q1': {'type':RGROLES.SINGLE, 'options' : {'0': {}, '1':{}}},
            'Q2': {'type':RGROLES.SINGLE, 'options' : {'0': {}, '1':{}}},
        })



from ...context import Context

from . import KNOWN_EXPRESSIONS
from typing import List,Union,Optional

ARG_SCALAR = 'scalar'
ARG_SURVEYKEY = 'survey_key'
ARG_ITEM_KEY = 'item_key'
ARG_RG_KEY = 'rg_key' # eg 'rg'
ARG_RG_ITEM_KEY = 'rg_item_key' # eg scg
ARG_RG_COMP_KEY = 'rg_comp_key' # eg '1'
ARG_RG_COMP_PREFIX = 'rg_comp_prefix' 
ARG_RG_COMP_FULL_PATH = 'rg_full_path' # eg full response key from item e.g. rg.scg.1
ARG_STUDY_STATUS = 'study_status'
ARG_QUALIFIED_RESPONSE ='qualified_response'
ARG_QUALIFIED_RG = 'qualified_rg'
ARG_FLAG_KEY = 'flag_key'
ARG_TIMESTAMP = 'timestamp'
ARG_PATH = 'path' # Path of objects like Q1.rg.scg.1
ARG_ITEM_PATH = 'item_path' # Path of any object from a item
ARG_VALIDATION_KEY = 'validation_key'
KNOWN_ARG_ROLES = [
    ARG_SCALAR,
    ARG_SURVEYKEY,
    ARG_ITEM_KEY,
    ARG_RG_KEY,
    ARG_RG_ITEM_KEY,
    ARG_RG_COMP_KEY,
    ARG_RG_COMP_PREFIX,
    ARG_RG_COMP_FULL_PATH,
    ARG_STUDY_STATUS,
    ARG_QUALIFIED_RESPONSE,
    ARG_QUALIFIED_RG,
    ARG_FLAG_KEY,
    ARG_TIMESTAMP,
    ARG_PATH,
    ARG_VALIDATION_KEY
]

class Arg:
    """
        Argument descripton
    """
    def __init__(self, name:str, pos: int, role:str=None, variadic:bool=False, allow_this:bool=False, optional:bool=False, description:str=None):
        self.name = name
        self.pos = pos
        self.variadic = variadic
        self.allow_this = allow_this
        self.optional = optional
        self.description = description
        self.role = role

    def __repr__(self) -> str:
        return "<Arg %d %s>" % (self.pos, self.name)

    def to_readable(self, ctx):
        return self.name

class CompositeArgument:
    def __init__(self, args:List[Arg]):
        self.args = args

    def __repr__(self) -> str:
        return self.args.__repr__()
    

class Reference:
    """
        Argument reference to another element in the survey definition.
        The reference type is defined by the role of the reference
    """
    def __init__(self, role: str):
        self.role = role

class KeyReference(Reference):
    """
        Reference to a key type in the survey
        Role defines the kind of key
    """
    def __init__(self, role: str, param:Arg):
        super(KeyReference, self).__init__(role)
        self.param = param

    def __repr__(self) -> str:
        return "<KeyRef %s %s>" % (self.role, self.param)

class ItemPathReference(Reference):
    """
        Reference to a key path of an item in the survey
        item_key: argument where the item key is defined
        paths : list of argument to construct the list
    """
    def __init__(self, role: str, item_key:Arg, path: Union[Arg, CompositeArgument] ):
        super(ItemPathReference, self).__init__(role)
        self.item_key = item_key
        self.path = path

    def __repr__(self) -> str:
        return '<ItemPath %s %s>' % (self.item_key, self.path)


class EnumerationReference(Reference):
    """
        Role with a predefined value
    """
    def __init__(self, role: str, param: Arg, values: List[str]):
        super(EnumerationReference, self).__init__(role)
        self.values = values
        self.param = param

    def __repr__(self) -> str:
        d = {"role": self.role, "param": self.param, "values": self.values}
        return d.__repr__()

class ArgList:

    def __init__(self, params: List[Arg]) -> None:
        self.params = params
        d = {}
        for p in params:
            if p.name in d: 
                raise Exception("Argument '%s' already in list" % (p.name,))
            d[p.name] = p
        self.names = d

    def __len__(self):
        return len(self.params)

    def get_by_name(self, name)->Optional[Arg]:
        if name in self.names:
            return self.names[name]
        return None

    def at(self, index)->Optional[Arg]:
        if index < 0:
            raise Exception("Out of bound index %d" % (index, ))
        if index >= len(self.params):
            raise Exception("Out of bound index %d" % (index, ))
        return self.params[index]

    def __getitem__(self, i):
        return self.at(i)

    def __iter__(self):
        return iter(self.params)

    def __repr__(self) -> str:
        return self.params.__repr__()

class BaseExpressionType:

    pass

class UnknownExpressionType(BaseExpressionType):
    """
        Expression only known by name (doesnt knows params)
    """
    def has_params(self):
        return False

    def has_refs(self):
        return False

    def __repr__(self) -> str:
        return '<UnknownExpr>'

class ExpressionType(BaseExpressionType):
    """
        Describe a type of expression
    """
    def __init__(self, params: Optional[ArgList]=None, references: Optional[List[Reference]]=None ):
        self.params = params
        self.references = references
        self.kind = None
        self.description = None

    def has_params(self):
        return self.params is not None and len(self.params) > 0
    
    def get_param(self, index):
        if self.params is None:
            return None
        if index > len(self.params)-1:
            return None
        return self.params[index]

    def has_refs(self):
        return self.references is not None and len(self.references) > 0

    def __repr__(self) -> str:
        d = {}
        if self.kind is not None:
            d['kind'] = self.kind
        if self.params is not None:
            d['params'] = self.params
        if self.references is not None:
            d['refs'] = self.references
        return '<Expr ' + d.__repr__() + '>'



"""
dictionary model


dictionary is a simplified view of the Influenzanet survey model centered on data collection and encoding

It's used to export as a simple structure (into 'readable' format) or to compare with the standard

"""
from typing import List,Optional

class OptionDictionnary:

    def __init__(self, key:str, role:str, item_key:str, rg_item: str, obj):
        self.key = key
        self.role = role
        self.item_key = item_key
        self.rg_item = rg_item
        self._obj = obj
    
    def __repr__(self):
        return self.to_readable().__repr__()

    def to_readable(self, ctx=None):
        """
            To readable representation (simple structure serializable as simple json or yaml)
        """
        return {
            'key': self.key, 
            'role': self.role, 
            'item_key': self.item_key,
            'rg_item': self.rg_item
        }
    
    def get_component(self):
        return self._obj

class ItemDictionnary:
    """
        Item dictionnary implements a simple question model from the Survey model, centered on data collection
        It only embeds information about data collection and encoding
    """
    def __init__(self, key:str, type:str, options:Optional[List[OptionDictionnary]], parent_key:str, obj):
        self.key = key
        self.type = type
        self.options = options
        self.parent_key = parent_key # Parent prefix
        self.rg_key = None # Response group key
        self._obj = obj

    def __repr__(self):
        return {'key': self.key, 'type': self.type}.__repr__()

    def to_readable(self, ctx=None):
        """
            Transforms the object into readable format
        """
        return {
            'key': self.key,
            'type': self.type,
            'options': self.options,
        }

    def get_survey_item(self):
        return self._obj



from typing import Dict, List, OrderedDict
from ..utils import read_json

# Survey version here is related to the survey-engine.ts version
# 1.1 = Survey with old history model (current and history fields, and SurveyItem with version,versionTags field) 
# 1.2 = Survey with only current (Survey doesnt embeds history anymore)

def read_survey_json(file):
    """
        Load a json survey and transform it to new format if needed
    """
    json = read_json(file)
    if 'studyKey' in json and 'survey' in json:
        return survey_transform_to_12(json)

    if 'surveyDefinition' in json and 'props' in json:
        return json
    else:
        raise Exception("The file '%s' doesnt seem to be a survey definition" % file)

def require_fields(json, fields):
    for field in fields:
        if not field in json:
            raise Exception("Missing field '%s'" % field)

def survey_transform_to_12(json, preserve_item_version=False):

    try:
        require_fields(json, ['survey', 'studyKey'])
    except Exception as e:
        raise Exception("Doest seems to be a survey version 1.1 %s " % str(e))

    def transform_item(item):
        if 'version' in item:
            if preserve_item_version:
                item["metadata"] = {'version': str(item['version'])}
            del item['version']
        if 'versionTags' in item:
            del item['versionTags']
        if 'items' in item:
            item['items'] = list(map(transform_item, item['items']))
        return item

    survey = json['survey'] # First level not used any more
    surveyVersion = survey['current']
    del survey['current']
    survey['surveyDefinition'] = transform_item(surveyVersion['surveyDefinition'])
    survey['versionId'] = surveyVersion['versionId']

    return survey

def survey_transform_to_11(json, study_key):
    """
        Take a survey dictionary from a Survey version 1.1 
    
    """
    try:
        require_fields(json, ['surveyDefinition', 'versionId'])
    except Exception as e:
        raise Exception("Doest seems to be a survey version 1.2 %s " % str(e))
    
    def transform_item(item):
        item['version'] = 1
        if 'items' in item:
            item['items'] = list(map(transform_item, item['items']))
        return reorder_dict(item, ['key','version', 'items'])
    
    survey_version = {
        'versionId': json['versionId'],
        'surveyDefinition': transform_item(json['surveyDefinition']),
    }
    del json['surveyDefinition']
    del json['versionId']
    
    if 'unpublished' in json:
        survey_version['unpublished'] = json['unpublished']
        del json['unpublished']
    
    if 'published' in json:
        survey_version['published'] = json['published']
        del json['published']
    
    json['current'] = survey_version

    json = reorder_dict(json, ['id','props', 'current','history','prefillRules','contextRules', 'maxItemsPerPage', 'availableFor','requireLoginBeforeSubmission'])
    return {
         "studyKey": study_key,
         "survey": json
    }


def reorder_dict(data: Dict, fields:List[str]):
    """
    Reorder a dictionnary fields with a list of field to put as first fields
    Fields in data not in field list are placed at the end
    """
    old_keys = list(data.keys())
    new_keys = []
    for field in fields:
        if field in old_keys:
           new_keys.append(field)
           old_keys.remove(field)
    new_keys += old_keys # Add old keys remaining at the end
    d = OrderedDict()
    for field in new_keys:
        d[field] = data[field]
    return d


import requests
import json
from time import sleep
from typing import List, Optional, Dict, Any
from requests.auth import HTTPBasicAuth

REPOSITORY_API = 'https://surveys.ifn.epipop.fr'
DEFAULT_NAMESPACE = 'influenzanet'

class ApiError(Exception):
    """ 
        Api Error can be used to return the error with status
    """
    def __init__(self, message, status=None, response=None):
        super(ApiError, self).__init__(message)
        self.status = status
        self.response = response       
    
    def is_too_many_request(self):
        """
            Check if response is 
        """
        return self.status is not None and self.status == 429
    
    def retry_after(self):
        """
            Get Retry-After header
        """
        if self.response is None or self.response.headers is None:
            return None
        retry = self.response.headers.get('retry-after')
        if retry is None:
            return None
        return int(retry)
    
    def wait_retry_after(self, show:bool=True):
        retry = self.retry_after()
        if retry is None:
            return False
        if show:
            print("Waiting {} secs from Retry-After header".format(retry))
        sleep(retry)
        return True
        
class ApiRateLimiter:
    """
        Collect information about an eventual rate limiter 
        Provides wait() method to wait, or wait_delay() to get the delay
    """
    def __init__(self, headers):
        limit = headers.get('x-ratelimit-limit', None)
        self.limit = 0
        self.remaining = 0
        self.reset = 0
        if limit is not None:
            self.limit = int(limit)
        remaining = headers.get('x-ratelimit-remaining', None)
        if remaining is not None:
            self.remaining = int(remaining)
        reset = headers.get('x-ratelimit-reset', None)
        if reset is not None:
            self.reset = int(reset) 
    
    def wait_delay(self):
        """
            Compute the delay to wait (in sec.) considering the rate limiter info
        """
        if self.limit == 0 or self.reset == 0:
            return 0
        if self.remaining == 0:
            return self.reset
        w = (self.remaining / self.reset) + 0.05
        return w

    def wait(self, show: bool=True):
        w = self.wait_delay()
        if w > 0:
            if show:
                print("Waiting %f seconds (limit=%d, remains=%d, reset=%d)" % (w, self.limit, self.remaining, self.reset))
            sleep(w)

class APIResponse:

    def __init__(self, data, headers):
        self.data = data
        self.headers = headers
        self.limiter = ApiRateLimiter(headers)

class ImportResponse:
    def __init__(self, meta, created:bool, limiter: ApiRateLimiter):
        self.id = meta.get('id', None)
        self.version = meta.get('version')
        self.published = meta.get('published')
        self.platform = meta.get('platform')
        self.survey_name = meta.get('name')
        self.model_type = meta.get('model_type')
        self.created = created
        self.limiter = limiter
 
class SurveyRepositoryAPI:

    def __init__(self, config: Optional[Dict[str, Any]]=None):
        """
        Create a API wrapper

        config is an optional dictionary.
        Can contain:
        `user`: user name to import survey
        `password`: password for user (only to import survey)
        `url`: customize repository url, only needed for dev
        `platform_code`: code of the platform to use for import 
        """
        if config is None:
            config = {}
        if not isinstance(config, dict):
            raise ValueError("config must be a dictionary")
        self.api_url = config.get('url', REPOSITORY_API)
        creds = None
        if 'user' in config:
            if 'password' not in config:
                raise ValueError("Missing field 'password' in survey repository config")
            creds = HTTPBasicAuth(config['user'], config['password'])
        self.credentials = creds
        self.platform_code = config.get('platform_code', None)

    def create_error(self, response: requests.Response):
        msg = None
        try:
            data = response.json()
            if isinstance(data, dict) and "error" in data:
                msg = data["error"]
        except:
            pass
        if msg is None:
            msg = "Request error : " + response.reason
        raise ApiError(msg, response.status_code, response=response)
        
    def import_survey(self, survey, platform=None, namespace=None, version=None, name=None):
        """
            Import survey definition into the repository

            Caution, it's possible to import a survey preview, but it must be one single version of a survey
            In this case, to provide the `name` parameter is mandatory
            survey: the survey data (serialized json as string)
            platform: the platform code from which the survey belongs
            credentials: the account credentials on the repository
            namespace: namespace of the repository, by default its influenzanet for the Influenzanet project.
            version: version id of the survey (if it's published in a platform),
            name: name of the survey, this is used when uploading a survey preview (survey key is not in the preview itself)
        """

        if platform is None:
            if self.platform_code is None:
                raise ValueError("platform code is not set in configuration, must be provided")
            platform = self.platform_code
        if namespace is None:
            namespace = DEFAULT_NAMESPACE
        url = "{}/import/{}".format(self.api_url, namespace)
        data = {
            'platform': platform,
        }
        if version is not None:
            data['version'] = version
        if name is not None:
            data['name'] = name
        
        files = {'survey': ('survey_json', survey, 'application/json')}
        r = requests.post(url, auth=self.credentials, files=files, data=data)
        if r.status_code >= 200 and r.status_code < 300:
            d = r.json()
            created = False
            if r.status_code == 201:
                created = True
            limiter = ApiRateLimiter(r.headers)
            response = ImportResponse(d, created=created, limiter=limiter)
            return response
        self.create_error(r)
        
    def list_surveys(self, namespace=None, platforms: Optional[List[str]]=None, names:Optional[List[str]]=None, types:Optional[List[str]]=None, limit:int=0, offset:int=0, short:bool=False):
        """
            List available surveys

            platforms: list of platforms code
            names: list of survey names
            types: list of model type ('P' for preview, 'D' for Definition)
        """
        if namespace is None:
            namespace = DEFAULT_NAMESPACE
        url = "{}/namespace/{}/surveys".format(self.api_url, namespace)
        if short:
            url += '/versions'
        params = {}
        if platforms is not None and len(platforms) > 0:
            params['platforms'] = ','.join(platforms)
        if names is not None and len(names) > 0:
            params['names'] = ','.join(names)
        
        if types is not None and len(types) > 0:
            params['types'] = ','.join(types)
        
        params['limit'] = limit
        if limit > 0:
            # Cannot provide offset if limit is 0
            params['offset'] = offset
        
        r = requests.get(url, params=params)
        if r.status_code == 200:
            return r.json()
        self.create_error(r)
        
    def load_survey_meta(self, id):
        """
            Load content of a survey definition for it's id
        """
        url = "{}/survey/{}".format(self.api_url, id)
        r = requests.get(url)
        if r.status_code == 200:
            return APIResponse(r.json(), r.headers)
        self.create_error(r)

    def load_survey_data(self, id):
        """
            Load content of a survey definition for it's id
        """
        #"/survey/:id/data"
        url = "{}/survey/{}/data".format(self.api_url, id)
        r = requests.get(url)
        if r.status_code == 200:
            return APIResponse(r.json(), r.headers)
        self.create_error(r)
       
         
        


from .model import Expression, Scalar

# Expression to object nodes
def expression_arg_parser(expr, level=0, idx=0):
    if not 'dtype' in expr:
        dtype = 'str'
    else:
        dtype = expr['dtype']
    
    if not dtype in expr:
        print("Invalid expression claim to be %s but no entry" % (dtype, ))
    value = expr[dtype]
    if dtype == 'exp':
        return expression_parser(value)
    return Scalar(dtype, value)

def expression_parser(expr, level=0, idx=0):
    params = []
    level = level + 1
    if 'data' in expr:
        idx = 0
        for d in expr['data']:
            p = expression_arg_parser(d, level, idx)
            params.append(p)
            idx = idx + 1
    if 'name' in expr:
        name = expr['name']
    else:
        name = "_%d" % (idx, )
    return Expression(name, params)    




from typing import Dict
from .translatable import to_translatable,parse_translatable
from .time import Timestamp
from .survey import Study, Survey, SurveyGroupItem, SurveyItemValidation, SurveySingleItem, SurveyItemGroupComponent, SurveyItemResponseComponent, SurveyItemComponent
from .expression import expression_arg_parser, expression_parser, Scalar

def component_parser(obj):
    role = obj['role']
    key =  obj.get('key')

    comp = None
    if 'items' in obj:
        # ItemGroupComponent
        ii = []
        for it in obj['items']:
            ii.append(component_parser(it))
        
        if 'order' in obj:
           order = expression_parser(obj['order'])
        else:
            order = None
        comp = SurveyItemGroupComponent(key=key, role=role, items=ii, order=order)

    if 'dtype' in obj:
        # ResponseComponent
        if comp is not None:
            raise Exception("Component cannot be group and response type")
        comp = SurveyItemResponseComponent(key=key, role=role, dtype=obj['dtype'])

    if comp is None:
        # ItemComponent base (Display ?)
        comp = SurveyItemComponent(key=key, role=role)

    if 'content' in obj:
        comp.content = parse_translatable(obj['content'])
    
    if 'properties' in obj:
            def parse_comp(value):
                if isinstance(value, str):
                    return Scalar("str", value)
                if isinstance(value, (int, float)):
                    return Scalar("num", value)
                return expression_arg_parser(value)
            pp = {}

            for k, p in obj['properties'].items():
                pp[k] = parse_comp(p)
            comp.properties = pp

    if 'description' in obj:
        comp.description = parse_translatable(obj['description'])

    if 'displayCondition' in obj:
        comp.displayCondition = expression_parser(obj['displayCondition'])

    if 'disabled' in obj:
        comp.disabled = expression_parser(obj['disabled'])

    if 'style' in obj:
        ss = {}
        for s in obj['style']:
            ss[s['key']] = s['value']
        comp.style = ss

    return comp

def survey_item_parser(obj):
    _id = obj.get('id')
    key = obj['key']
    if 'items' in obj:

        ii = []
        for i in obj['items']:
            ip = survey_item_parser(i)
            ii.append(ip)
        if 'selectionMethod' in obj:
            selection = expression_parser(obj['selectionMethod'])
        else:
            selection = None
        item = SurveyGroupItem(key,  id=_id, items=ii, selection=selection)

    else:
        comp = None
        if 'components' in obj:
            comp = component_parser(obj['components'])
        _type = obj.get('type')
        if 'validations' in obj:
            vv = []
            for v in  obj['validations']:
                rule = expression_parser(v['rule'])
                validation = SurveyItemValidation(v['key'], v['type'], rule) 
                vv.append(validation)
            validations = vv
        else:
            validations = None
        item = SurveySingleItem(key, id=_id, type=_type, components=comp, validations=validations)

    if 'condition' in obj:
        condition = expression_parser(obj['condition'])
        item.condition = condition
    if 'follows' in obj:
        item.follows = obj['follows']
    
    if 'priority' in obj:
        item.priority = obj['priority']

    if 'metadata' in obj:
        item.setMetadata(obj['metadata'])

    return item

def context_rules_parser(ctx: Dict):
    if 'mode' in ctx:
        ctx['mode'] = expression_arg_parser(ctx['mode'])
    if 'previousResponses' in ctx:
        rr = []
        for r in ctx['previousResponses']:
            exp = expression_parser(r)
            rr.append(exp)
        ctx['previousResponses'] = rr
    return ctx


def survey_parser(survey:Dict):
    """
    Parse a dictionary based survey (loaded from a json definition file) model into a python object model
    
    Returns: Survey

    """
    pp =  survey['props']
    pp = to_translatable(pp, ['name','description', 'typicalDuration'])
    survey['props'] = pp
    if 'published' in survey:
        # published is not present when survey is only a draft from editor
        survey['published'] = Timestamp(survey['published'])
    if 'unpublished' in survey:
        survey['unpublished'] = Timestamp(survey['unpublished'])
    survey['surveyDefinition'] = survey_item_parser(survey['surveyDefinition'])

    if 'prefillRules' in survey:
        rr = []
        for r in survey['prefillRules']:
            exp = expression_parser(r)
            rr.append(exp)
        survey['prefillRules'] = rr

    if 'contextRules' in survey:
        survey['contextRules'] = context_rules_parser(survey['contextRules'])

    return Survey(survey)

def study_parser(study):
    """"
     Parse study to have some elements easier to represents (like expression)
    """
    if 'rules' in study:
        # Replace
        rr = []
        for rule in study['rules']:
            r = expression_parser(rule)
            rr.append(r)
        study['rules'] = rr
    if 'props' in study:
        pp = study['props']
        pp = to_translatable(pp, ['name','description'])
        if 'tags' in pp:
            tt = []
            for tag in pp['tags']:
                if 'label' in tag:
                    tag['label'] = parse_translatable(tag['label'])
                tt.append(tag)
            pp['tags'] = tt
        study['props'] = pp
    return Study(study)




import os
import json
from typing import Dict

from . import KNOWN_EXPRESSIONS,KNOWN_ENUMS, library_path

from .types import *

class ParserException(Exception):
    pass

def load_library():
    path = library_path()
    parser = ExpressionTypeParser()
    parser.parse(path, KNOWN_EXPRESSIONS, KNOWN_ENUMS)
    return (KNOWN_EXPRESSIONS, KNOWN_ENUMS)

class ExpressionTypeParser:

    def __init__(self) -> None:
        self.enums = None

    def parse(self, path, knows: Dict[str, ExpressionType], enums: Dict[str, List] ):
        self.enums = enums
        try:
            data = json.load(open(path, 'r', encoding='UTF-8'))
        except Exception as e:
            raise ParserException("Unable to load json in %s" % path) from e

        for name, values in data['enums'].items():
            self.enums[name] = values

        defs = data['expressions']
        for name, expDef in defs.items():
            try:
                exp = self.parse_exp_definition(expDef)
            except Exception as e:
                raise ParserException("Error parsing '%s'" % name) from e
            knows[name] = exp

    def parse_arguments(self, args: List, refs:List)->ArgList:
        params = []
        role = None
        for index, a in enumerate(args):
            variadic = False
            role = None

            extra = {}

            if isinstance(a, str):
                # We only get name
                name = a
            else:
                # Object description
                if 'variadic' in a:
                    extra['variadic'] = a['variadic']
                if 'allow_this' in a:
                    extra['allow_this'] = a['allow_this']
                if 'optional' in a:
                    extra['optional'] = a['optional']
                if 'description' in a:
                    extra['description'] = a ['description']
                name = a['name']
                if 'role' in a:
                    role = a['role']
            p = Arg(name, pos=index, role=role, **extra)
            
            if role is not None:
                if role in self.enums:
                    ref = EnumerationReference(role, p, self.enums[role])
                else:
                    if not role in KNOWN_ARG_ROLES:
                        raise Exception("Unknown role '%s'" % (role, ))
                    ref = KeyReference(role, p)
                refs.append(ref)
            params.append(p)
        return ArgList(params)
    
    def parse_exp_definition(self, expDef):
        params = None
        roles = [] # References
        if isinstance(expDef, list):
            # This is a argument list
            params = self.parse_arguments(expDef, roles)
        else:
            # Object describing the function
            if 'params' in expDef:
                params = self.parse_arguments(expDef['params'], roles)
            else:
                raise Exception("Unknown expression type, missing 'params' entry")
            if 'roles' in expDef:
                for r in expDef['roles']:
                    role = self.parse_role(r, params)
                    roles.append(role)
        if len(roles) == 0:
            roles = None
        exp_type=  ExpressionType(params, roles)
        if 'description' in expDef:
            exp_type.description = expDef['description']
        return exp_type

    def parse_role(self, roleDef, params:ArgList):
        """
            Parse a role definition in the "roles" entry or an expression definition
            Those definitions can create roles combining several arguments 
        """

        def get_arg(name):
            arg = params.get_by_name(name)
            if arg is None:
                raise Exception("Unknown argument '%s'" % (name))
            return arg

        def get_arg_list(names):
            pp = []
            for p in names:
                arg = get_arg(p)
                pp.append(arg)
            return pp

        role = roleDef['role']

        if role == ARG_ITEM_PATH:
            item_arg = get_arg(roleDef['params']['item_key'])
            path_args = get_arg_list(roleDef['params']['path'])
            if len(path_args) == 1:
                path_args = path_args[0]
            else:
                path_args = CompositeArgument(path_args)

            return ItemPathReference(role, item_arg, path_args)

        return None
        

def render_library(funcs: Dict[str, ExpressionType])->str:
    output = [
        "# Survey Engine Expressions"
    ]
    for name, func in funcs.items():
        out = render_func(name, func)
        output.append("")
        output.append(out)
    return "\n".join(output)

ROLE_NAMES = {
    ARG_FLAG_KEY:"Flag name",
    ARG_ITEM_KEY:"Survey item key",
    ARG_SURVEYKEY:"Survey key",
    ARG_STUDY_STATUS:"Study status name",
    ARG_ITEM_PATH:"Component path of an item",
    ARG_RG_COMP_KEY:"Individual key of a component",
    ARG_RG_COMP_PREFIX:"Prefix of a response group component, e.g. 'rg.scg.', 'rg.mcg.'",
    ARG_RG_KEY:"Key of a response group, e.g. 'rg'",
    ARG_TIMESTAMP:"timestamp",
    ARG_VALIDATION_KEY:"Key of a 'validations' rule of a component"
}

def render_func(name:str, func: ExpressionType)->str:
    o = [
        "## " + name
    ]

    if isinstance(func, UnknownExpressionType):
       return "\n".join(o)

    if func.kind == "action":
        o.append("Survey **action**, useable in survey rules")
    if func.kind == "client":
        o.append("Client side function")
    if func.kind == "service":
        o.append("Service side function, useable in survey rules")

    if func.has_params():
        o.append("")
        o.append("** Parameters **")
        o.append("")
        for p in func.params:
            d = "- `%s`" % (p.name, )
            if p.variadic:
                d += "... " 
            if p.description is not None:
                d += " " + p.description
            else:
                if p.role in ROLE_NAMES:
                    d += " " + ROLE_NAMES[p.role] 
            if p.allow_this:
                d += " [accepts `this` as special value]"
            if p.optional:
                d += " **optional**"
            o.append(d)
    return "\n".join(o)



from .expression import Expression, expression_parser

from typing import List, Union

TranslatePart = Union[str, Expression]

class Translatable:
    
    def __init__(self, code:str, value):
        """
        docstring
        """
        self.code = code
        self.value = value

    def value_as_text(self, context):
        if isinstance(self.value, list):
            vv = []
            for item in self.value:
                if(isinstance(item, Expression)):
                    v = item.to_readable(context)
                    if not isinstance(v, str):
                        v = str(v)
                else:
                    v = str(item)
                vv.append(v)
            return ' '.join(vv)

    def get_parts(self) -> List[TranslatePart]:
        return self.value

    def get_code(self)->str: 
        return self.code

    def __repr__(self):
        return "<T[%s, %s]>" % (self.code, self.value)
        
class TranslatableList:

    def __init__(self, values:List[Translatable]):
        self.values = values

    def get_with_context(self, context)->List[Translatable]:
        language = context.get_language()
        if language is None:
            return self.get_translates()
        tt = []
        if isinstance(language, str):
            language = [language]

        for t in self.values:
            if t.code in language or t.code.startswith('_'):
                tt.append(t)
        return tt

    def get_translates(self)->List[Translatable]:
        return self.values

    def to_readable(self, ctx):
        render_translatable(self, ctx)


def to_translatable(data, fields):
    """
        Transform some fields of a dictionnary to translatable object
    """
    for field in fields:
        if field in data:
            data[field] = parse_translatable(data[field])
    return data

def parse_translatable(values):
    """
        Parse a translatable data structure {code:, parts: [...]} to an object based structure
        Returns TranslatableList
    """
    tt = []
    for value in values:
        code = value['code']
        texts = []
        for p in value['parts']:
            if 'str' in p:
                texts.append(p['str'])
                break
            if 'num' in p:
                texts.append(str(p['num']))
            if 'exp' in p:
                exp = expression_parser(p['exp'])
                texts.append(exp)

        tt.append(Translatable(code, texts))
    return TranslatableList(tt)

def render_translatable(t, context):
    """
        Render a Translatable list object
    """
    trans_to_render = t.get_with_context(context)
    values = []

    several = len(trans_to_render) > 1
    for t in trans_to_render:
        if several:
            text = "[%s] %s" % (t.code, t.value_as_text(context))
        else:
            text = t.value_as_text(context)
        values.append(text)
    if len(values) == 0:
        return ''
    if len(values) == 1:
        return values[0]
    return values 

def readable_translatable(values, context):
   tt = parse_translatable(values)
   return render_translatable(tt, context)




from .expression import *
from .translatable import *
from .time import *
from .dictionnary import *
from .survey import *
from .responses import *
from .parser import *
from .render import *
from .loader import *


from collections import OrderedDict
import logging
from inspect import getframeinfo, stack
from ...influenzanet.translatable import TranslatableList
from ...influenzanet.survey import SurveyPath
from ...influenzanet.expression.types import ARG_ITEM_KEY, ARG_SURVEYKEY, Arg, CompositeArgument, EnumerationReference, ItemPathReference, KeyReference, UnknownExpressionType
from ...influenzanet.expression import KNOWN_EXPRESSIONS, ExpressionType, find_expression_type
from ...influenzanet.expression.library import load_library

from ...influenzanet import Study, Survey, SurveyItem, SurveyItemComponent, Expression, OptionDictionnary, RGROLES
from ...context import Context

from typing import List, Optional,Dict, Union

logger = logging.getLogger(__name__)

def get_arg_value(arg:Arg, exp:Expression):
    """
        Get passed values for arguments
    
    """
    max_pos = len(exp.params)-1
    if max_pos < arg.pos:
        return None
    if arg.variadic:
        p =  exp.params[arg.pos:]
        if len(p) == 1:
            return p[0]
        return p
    return exp.params[arg.pos]

class CheckContext(dict):

    def __init__(self, parent=None, **kwargs):
        super(CheckContext, self).__init__(**kwargs)
        self.parent = parent
        self.caller = getframeinfo(stack()[1][0])
    
    def to_readable(self, ctx):
        """
            Create a readable form of context as a list (parents first) of key values
        """
        pp = []
        self.parents(pp)
        d = []
        pp.reverse()
        for p in pp:
            kw = []
            for k,v in p.items():
                kw.append("%s=%s" % (k,v))
            line = "%s at %d" % (','.join(kw), p.caller.lineno)
            d.append(line)
        return d

    def parents(self, stack:List):
        """
            Walk up into parents hierarchy
        """
        stack.append(self)
        if self.parent is not None:
            self.parent.parents(stack)

class Problem:
    """
        Describe a problem
    """
    DUPLICATE_KEY = 'dup_key'
    DUPLICATE_RESPONSE_KEY = 'dup_response_key'
    UNKNOWN_EXP = 'unknown_exp'
    UNCHECKABLE = 'uncheckable'
    UNKNOWN_REF = 'unknown_ref'

    def __init__(self, pb_type, value=None, ctx:CheckContext=None, line=None):
        self.type = pb_type
        self.value = value
        self.line = line
        self.ctx = ctx

    def to_readable(self, ctx: Context):
        d = {'type': self.type,}
        if self.line is not None:
            d['raised_at'] = self.line
        if self.value is not None:
            d['value'] = self.value
        d['context'] = self.ctx
        return d

class SurveyChecker:

    LOADED = False

    def __init__(self):

        if not self.LOADED:
            load_library()
            self.LOADED = True
        self.item_keys = OrderedDict()
        self.known_surveys = set()
        self.problems = []

    def check(self, survey: Survey):
        
        definition = survey.survey_definition
    
        self.known_surveys.add(definition.key)

        self.discover(definition)

        ctx = CheckContext(survey="current")
        self.check_item(definition, ctx)

        return self.problems

    def notify(self, pb_type, ctx:CheckContext, value=None):
        caller = getframeinfo(stack()[1][0])
        pb = Problem(pb_type, ctx=ctx, value=value, line=caller.lineno)
        self.problems.append(pb)

    def discover(self, survey: SurveyItem):
        """
            Discover create data about survey to supports check of unicity of keys
        """
        dd = survey.get_dictionnary()
        for item in dd:
            ctx = CheckContext(item=item.key)
            if item.key in self.item_keys:
                self.notify(Problem.DUPLICATE_KEY, ctx)
            else:
                self.item_keys[item.key] = item
            if item.options is not None:
                if item.type in [RGROLES.SINGLE, RGROLES.DROPDOWN, RGROLES.LIKERT, RGROLES.MULTIPLE]:
                    self.check_options(item.options, ctx)
                    
    def check_options(self, options:List[OptionDictionnary], parent:CheckContext):
        kk = dict()
        for o in options:
            logger.debug("check option %s" % (o.key))
            if o.key in kk:
               ctx = CheckContext(parent=parent, key=o.key)
               self.notify(Problem.DUPLICATE_RESPONSE_KEY, ctx=ctx)
            else:
                kk[o.key] = True
        
    def check_item(self, surveyItem: SurveyItem, parent: CheckContext):
        logger.debug("check item %s" % (surveyItem.key))
        
        if surveyItem.condition is not None:
            logger.debug("check item %s field 'condition'" % (surveyItem.key))
            ctx = CheckContext(item=surveyItem.key, field="condition", parent=parent)
            self.check_expression(surveyItem.condition, ctx)
        
        if surveyItem.is_group():
            # GroupItem
            for item in surveyItem.items:
                self.check_item(item, parent=parent)
            
            if surveyItem.selection is not None:
                ctx = CheckContext(item=item.key, field="selection", parent=parent)
                self.check_expression(surveyItem.selection, ctx)
        else:
            # SingleItem
            if surveyItem.validations is not None:
                for index, validation in enumerate(surveyItem.validations):
                    logger.debug("check item %s field 'validation' '%s'" % (surveyItem.key, validation.key))
                    ctx = CheckContext(parent=parent, item=surveyItem.key, field="validations", index=index, key=validation.key)
                    self.check_expression(validation.rule, ctx)
            ctx = CheckContext(parent=parent, item=surveyItem.key, field='components')
            self.check_component(surveyItem.components, ctx)
    
    def check_component(self, component:SurveyItemComponent, parent:CheckContext):
        logger.debug("check component '%s'" % (component.key))
        fields = ['displayCondition','disabled']
        if component.is_group():
            for index, comp in enumerate(component.items):
                ctx = CheckContext(parent=parent, key=component.key, field="items", index=index)
                self.check_component(comp, ctx)
            fields.append('order')
        
        for field in fields:
            logger.debug("check component field '%s' of '%s'" % (field, component.key))
            if hasattr(component, field) and getattr(component, field) is not None:
                ctx = CheckContext(parent=parent, key=component.key, field=field)
                self.check_expression(getattr(component, field), ctx)
        
        props = getattr(component, 'properties', None)
        if props is not None:
            for name, expr in props.items():
                logger.debug("check component prop '%s' of '%s'" % (name, component.key))
                ctx = CheckContext(parent=parent, key=component.key, field="properties", name=name)
                self.check_expression(expr, ctx)

        tt = ['content','description']
        for field in tt:
            if hasattr(component, field):
                logger.debug("check component field '%s' of '%s'" % (field, component.key))
                ctx = CheckContext(parent=parent, field=field)
                self.check_localized(getattr(component, field), parent=ctx)

    def check_localized(self, localized, parent: CheckContext):
        if isinstance(localized, TranslatableList):
            translates = localized.get_translates()
        else:
            translates = [localized]
        for t in translates:
            for index, part in enumerate(t.get_parts()):
                if isinstance(part, Expression):
                    ctx = CheckContext(parent=parent, index=index, code=t.get_code())
                    self.check_expression(part, ctx)

    def check_expression(self, exp: Expression, parent:CheckContext):
        logger.debug("check expr %s" % (exp))
        if exp.is_scalar():
            return
        context = CheckContext(parent=parent, name=exp.name)
        if exp.is_expression():
            has_params = len(exp.params) > 0
            exp_type = find_expression_type(exp.name)
            if exp_type is None or isinstance(exp_type, UnknownExpressionType):
                self.notify(Problem.UNKNOWN_EXP, context, exp.name)
            else:
                if exp_type.has_refs() and has_params:
                    self.check_expression_refs(exp, exp_type, context)
            if not has_params:
                return
            for p in exp.params:
                if p.is_expression():
                    self.check_expression(p, context)

    def check_expression_refs(self, exp:Expression, exp_type:ExpressionType, parent: CheckContext):
       for ref in exp_type.references:
           if isinstance(ref, EnumerationReference):
               self.check_enumeration_ref(exp, ref, parent)
           if isinstance(ref, KeyReference):
               self.check_key_reference(exp, ref, parent)
           if isinstance(ref, ItemPathReference):
               self.check_path_reference(exp, ref, parent)

    def check_enumeration_ref(self, exp:Expression, ref:EnumerationReference, parent: CheckContext):
        logger.debug("check enumeration %s" % (ref))
        ctx = CheckContext(parent=parent, role=ref.role, param=ref.param)
        p = get_arg_value(ref.param, exp)
        if p is None:
            return
        if isinstance(p, Expression):
            self.notify(Problem.UNCHECKABLE, ctx)
            return
        # It's a scalar
        if not p.value in ref.values:
            self.notify(Problem.UNEXPECTED_VALUE, ctx)

    def check_key_reference(self, exp:Expression, ref:KeyReference, parent: CheckContext):
        logger.debug("check key %s in %s" % (ref, exp))
        ctx = CheckContext(parent=parent, role=ref.role, param=ref.param)
        p = get_arg_value(ref.param, exp)
        if p is None:
            return
        if isinstance(p, list):
            values = p
        else:
            values = [ p ]
        for v in values:
            if isinstance(v, Expression):
                self.notify(Problem.UNCHECKABLE, ctx)
                continue
            value = str(v)
            if ref.role == ARG_ITEM_KEY:
                if value =="this" and ref.param.allow_this:
                    logger.debug("Item key is 'this', skipping")
                    continue
                if not value in self.item_keys:
                    self.notify(Problem.UNKNOWN_REF, CheckContext(parent=ctx, value=value))
                    continue
                logger.debug("Item key '%s' found" % value)
                continue
            if ref.role == ARG_SURVEYKEY:
                if not value in self.known_surveys:
                    self.notify(Problem.UNKNOWN_REF, CheckContext(parent=ctx, value=value))
                logger.debug("Survey key '%s' found" % value)
        
    def check_path_reference(self, exp:Expression, ref:ItemPathReference, parent: CheckContext):
        logger.debug("Check path %s in %s" % (ref, exp))
        item_key = get_arg_value(ref.item_key, exp)
        
        if item_key is None:
            logger.debug("Item %s not found %s" % (item_key))
            # Cannot check
            return
        
        context = CheckContext(parent=parent, param=ref.item_key)
        if isinstance(item_key, Expression):
            self.notify(Problem.UNCHECKABLE, context)
            return

        item_key = str(item_key)
        
        if item_key in self.item_keys:
            item = self.item_keys[item_key].get_survey_item()
        else:
            logger.debug("Unable to find item '%s'" % (item_key))
            self.notify(Problem.UNKNOWN_REF, context, value=item_key)
            # Should have been already notified by the argument check ?
            return

        context = CheckContext(parent=parent, item=item_key)

        args = []

        if isinstance(ref.path, CompositeArgument):
            args = ref.path.args
        else:
            args = [ref.path]
        
        stack = [] # Stack of current params values
        paths = [] # List of paths to check
        last = len(args) - 1
        for index,arg in enumerate(args):
            arg_value = get_arg_value(arg, exp)
            if arg_value is None:
                logger.debug("Path %d  %s not found" % (index, arg))
                continue
            if isinstance(arg_value, Expression):
                ctx = CheckContext(parent=context, param=index)
                self.notify(Problem.UNCHECKABLE, ctx)
                return
            if index == last:
                if arg.variadic and isinstance(arg_value, list):
                    for v in arg_value:
                        p = stack[:] # Get previous args values in stack
                        p.append(str(v)) # Add the current element
                        paths.append(SurveyPath(p)) # And create the path
                else:
                    stack.append(str(arg_value))
                    paths.append(SurveyPath(stack))
            else:
                stack.append(str(arg_value))

        if len(paths) == 0:
            logger.debug("No path in expression")
        
        # Currently consider all path segments doesnt contains '.' as it's path separator (SurveyItem keys are an exception)
        for path in paths:
            obj = item.get_in_path(path)
            p = '/'.join(path.traversed())
            if obj is None:
                self.notify(Problem.UNKNOWN_REF, context, value=path)
                continue
            logger.debug("Item path found %s in %s" % (p, item_key))

    

            
               
                


                   

        
        
    



KNOWN_EXPRESSIONS = {}
KNOWN_ENUMS = {}
import os

from typing import Optional

from .model import *
from .parser import expression_parser, expression_arg_parser
from .readable import readable_expression
from .types import ExpressionType

def find_expression_type(name:str)->Optional[ExpressionType]:
    if name in KNOWN_EXPRESSIONS:
        return KNOWN_EXPRESSIONS[name]
    return None

def library_path():
    return os.path.join(os.path.dirname(__file__), 'expressions.json')

def schema_path():
    return os.path.join(os.path.dirname(__file__), 'expressions.schema')


from typing import List, Optional, Dict, Union

from .responses import RG_ROLES, RG_ROLES_DATA
from .dictionnary import ItemDictionnary, OptionDictionnary
from .expression import Expression

TYPE_PAGE_BREAK = 'pageBreak'
ROLE_RESPONSE_GROUP = 'responseGroup'
ROLE_TEXT = 'text'
DISPLAY_ROLES = [ROLE_TEXT, 'label']
TYPE_SURVEY_END = 'surveyEnd'

###
#  SurveyItem # Abstract
#  SurveyGroupItem(SurveyItem)
#    items: [SurveyItem, SurveyGroupItem, SurveySingleItem]
#  SurveySingleItem(SurveyItem):
#     role:
#     components:  List[SurveyItemComponent]
#  SurveyItemGroupComponent(SurveyItemComponent)
#    items: [SurveyItemComponent, SurveyItemGroupComponent, SurveyItemResponseComponent]
#  SurveyItemComponent:
#    role
#
#

class SurveyPath:
    """
        Path traversal into survey node
    
    """
    def __init__(self, paths: List):
        pp = []
        for p in paths:
            if '.' in p:
                for sp in p.split('.'):
                    pp.append(sp)
            else:
                pp.append(p)
        self.paths = pp
        self.current = 0 # 

    def pop(self):
        if self.empty():
            return None
        p = self.paths[self.current]
        self.current += 1
        return p

    def is_last(self):
        """
            One least node
        """
        return self.current == len(self.paths) - 1

    def empty(self):
        """
            No more node to traverse
        """
        return self.current > len(self.paths) - 1

    def traversed(self):
        if self.empty():
            return self.paths
        if len(self.paths) == 0:
            return []
        if self.current == 0:
            return []
        return self.paths[0:self.current-1]

    def __len__(self):
        return len(self.paths)

    def __repr__(self) -> str:
        return {'paths':self.paths, 'current': self.current  }.__repr__()

    def to_readable(self, ctx):
        return {'paths': self.paths, 'current': self.current}
        
class SurveyItemComponent:
    
    def __init__(self, key, role ):
        self.key = key
        self.role = role
        
    def get_readable_label(self, name):
        if self.key is not None and self.key != '':
            k = "key=%s, role=%s" % (self.key, self.role)
        else:
            k = str(self.role)
        label = "%s<role=%s>" % (name, k, )
        return label

    def get_common_fields(self, o):
        for a in ['content', 'description', 'disabled', 'displayCondition','style', 'properties']:
            v = getattr(self, a, None)
            if v is not None:
                o[a] = v

    def to_readable(self, ctx):
        o = {
            '_ref': self.get_readable_label('DisplayComponent')
        }
        self.get_common_fields(o)
        return o

    def is_group(self):
        return False

    def is_response(self):
        return False

    def is_base(self):
        return True

    def get_type(self):
        return 'base'

    def get_in_path(self, path:SurveyPath):
        if not path.is_last():
            # No sub component, path cannot be found
            return None
        key = path.pop()
        if self.key == key:
            return self
        return None
       
class SurveyItemGroupComponent(SurveyItemComponent):
    
    def __init__(self, key, role, items, order):
        super(SurveyItemGroupComponent, self).__init__(key=key, role=role)
        self.items = items
        self.order = order
        
    def to_readable(self, ctx):
        o = {
            '_ref': self.get_readable_label('GroupComponent'),
            'items': self.items,
            'order': self.order
        }
        self.get_common_fields(o)
        return o

    def items_by_role(self, role)->Optional[List[SurveyItemComponent]]:
        if self.items is None:
            return None
        ii = []
        for item in self.items:
            if item.role == role:
                ii.append(item)
        return ii

    def items_by_roles(self)->Optional[Dict[str, List[SurveyItemComponent]]]:
        ## Group items by roles
        if self.items is None:
            return None
        roles = {}
        for item in self.items:
            r = item.role
            if not r in roles:
                roles[r] = []
            roles[r].append(item)
        return roles 
    
    def get_in_path(self, path: SurveyPath):
        """
            Check 
        """
        if self.items is None:
            return None
        key = path.pop()
        for item in self.items:
            if key == item.key:
                if not path.empty():
                    return item.get_in_path(path)
                return item
        return None

    def is_group(self):
        return True

    def is_response(self):
        return False

    def is_base(self):
        return False

    def get_type(self):
        return 'group'

class SurveyItemResponseComponent(SurveyItemComponent):
    
    def __init__(self, key, role, dtype):
        super(SurveyItemResponseComponent, self).__init__(key=key, role=role)
        self.dtype = dtype
    
    def to_readable(self):
        o = {
            '_ref': self.get_readable_label('ResponseComponent'),
            'dtype': self.dtype,
        }
        self.get_common_fields(o)
        return o

    def is_group(self):
        return False

    def is_response(self):
        return True

    def is_base(self):
        return False

    def get_type(self):
        return 'response'
        
class SurveyItem:
    
    def __init__(self, key, id=None):
        self.key = key
        self.id = id
        self.condition:Expression = None
        self.follows:List[str] = None 
        self.priority = None
        self.metadata: Dict = {}

    def setMetadata(self, metadata: Dict):
        self.metadata = metadata

    def get_readable_label(self, name):
        if self.id is not None:
            k = "key=%s, id=%s" % (self.key, self.id)
        else:
            k = str(self.key)
        label = "%s<key=%s>" % (name, k, )
        return label

    def get_dictionnary(self, parent_key:Optional[str]=None)-> Optional[List[ItemDictionnary]]:
        """
        Get flat list of data elements
        """
        return None

    def is_group(self):
        """
        is Item a group Item (with sub items)
        """
        return False
   
    def flatten(self):
        yield self


class SurveyItemValidation:

    def __init__(self, key:str, type: str, rule:Union[Expression, bool]) -> None:
        self.key = key
        self.type = type
        self.rule = rule

class SurveySingleItem(SurveyItem):

    def __init__(self, key, components: Optional[SurveyItemGroupComponent], validations: Optional[List[SurveyItemValidation]], type, id=None):
        super(SurveySingleItem, self).__init__(key=key, id=id)
        self.components = components
        self.validations = validations
        self.type = type

    def to_readable(self, ctx):
        o = {
            '_ref': self.get_readable_label('SingleItem'),
        }
        if self.type is not None:
            o['type'] = self.type
        o['components'] = self.components
        if self.validations is not None:
            o['validations'] = self.validations
        return o

    def get_dictionnary(self, parent_key:Optional[str]=None)-> Optional[ItemDictionnary]:
        rg = self.get_response_group()
        
        if self.type == TYPE_SURVEY_END:
            return None

        if rg is None:
            return None
        
        if len(rg) > 1:
            raise Exception("Several response group for %s "  % str(self) )
        
        if len(rg) == 0:
            print("Warning no response group for %s" % str(self) )
                
        if len(rg) == 1:
            rg = rg[0]
            # print("ResponseGroup of %s %s" %  (self.key, type(rg)))
            for rg_item in rg.items:
                role = rg_item.role
                if not role in RG_ROLES:
                    print("Warning unknown role %s" % (role, ))
                # Find the component item with options
                oo = None
                if rg_item.is_group():
                    # If it's a group let's find options
                    oo = self._get_response_options(rg_item)
                d = ItemDictionnary(self.key, role, oo, parent_key, self)
                d.rg_key = rg.key
                return d
        return None  
            
    def _get_response_options(self, itemComponent:SurveyItemComponent, root_key=None)->List[OptionDictionnary]:
        key = itemComponent.key
        if key is None:
            return []
        if root_key is not None:
            key = root_key + '.' + itemComponent.key
        options = []
        for item in itemComponent.items:
            if item.role in DISPLAY_ROLES:
                continue
            if item.key is None:
                continue
            if item.is_group():
                options.extend( self._get_response_options(item, key) )
            else:
                options.append(
                    OptionDictionnary(key + '.' + item.key, item.role, item.key, itemComponent.key, obj=item)
                )
        return options    

    def get_response_group(self)->Optional[List[SurveyItemComponent]]:
        if self.components is None:
            return None

        if self.type == TYPE_PAGE_BREAK:
            # No response for page break
            return None

        return self.components.items_by_role(ROLE_RESPONSE_GROUP)

    def get_in_path(self, path:SurveyPath):
        if self.components is None:
            return None
        return self.components.get_in_path(path)

    def __str__(self):
        return '<SurveySingleItem key=%s, type=%s>' % (self.key, self.type)

class SurveyGroupItem(SurveyItem):

    def __init__(self, key, items, selection, id=None):
        super(SurveyGroupItem, self).__init__(key=key, id=id)
        self.items = items
        self.selection = selection

    def to_readable(self, ctx):
        return {
            '_ref': self.get_readable_label('GroupsItem'),
            'items': self.items,
            'selection': self.selection
        }

    def get_dictionnary(self, parent_key: Optional[str]=None)-> Optional[List[ItemDictionnary]]:
        d = []
        parent_key = self.key
        for item in self.items:
            item_dict = item.get_dictionnary(parent_key)
            if item_dict is None:
                continue
            if isinstance(item_dict, list):
                d.extend(item_dict)
            else:
                d.append(item_dict)
        return d

    def __str__(self):
        return '<SurveyGroupItem %s, %s>' % (self.key, str(self.items))

    def is_group(self):
        """
        is Item a group Item (with sub items)
        """
        return True

    def flatten(self):
        yield self
        for item in self.items:
            yield from item.flatten()
class Study(dict):
    pass


class Survey(dict):
    
    def get_name(self):
        return self['props']['name']
        
    def getCurrent(self)->SurveyItem:
        """
            Get the current version definition of the survey
            Deprecated use .survey_definition instead
        """
        return self['surveyDefinition']

    @property
    def survey_definition(self)->SurveyItem:
        return self['surveyDefinition']

    @property
    def metadata(self)->Optional[Dict]:
        """
            get Survey metadata
        """
        if 'metadata' in self:
            return self['metadata']
        return None

    @property
    def version_id(self):
        """
            get survey version Id
        """
        if 'versionId' in self:
            return self['versionId']
        return None


from typing import Dict,List, Optional

EMBEDDED_CLOZE_PREFIX = 'embedded_cloze_'

QUESTION_TYPE_CONSENT                         = "consent"
QUESTION_TYPE_SINGLE_CHOICE                   = "single_choice"
QUESTION_TYPE_MULTIPLE_CHOICE                 = "multiple_choice"
QUESTION_TYPE_TEXT_INPUT                      = "text"
QUESTION_TYPE_NUMBER_INPUT                    = "number"
QUESTION_TYPE_DATE_INPUT                      = "date"
QUESTION_TYPE_DROPDOWN                        = "dropdown"
QUESTION_TYPE_LIKERT                          = "likert"
QUESTION_TYPE_LIKERT_GROUP                    = "likert_group"
QUESTION_TYPE_EQ5D_SLIDER                     = "eq5d_slider"
QUESTION_TYPE_NUMERIC_SLIDER                  = "slider"
QUESTION_TYPE_RESPONSIVE_TABLE                = "responsive_table"
QUESTION_TYPE_MATRIX                          = "matrix"
QUESTION_TYPE_MATRIX_RADIO_ROW                = "matrix_radio_row"
QUESTION_TYPE_MATRIX_DROPDOWN                 = "matrix_dropdown"
QUESTION_TYPE_MATRIX_INPUT                    = "matrix_input"
QUESTION_TYPE_MATRIX_NUMBER_INPUT             = "matrix_number_input"
QUESTION_TYPE_MATRIX_CHECKBOX                 = "matrix_checkbox"
QUESTION_TYPE_RESPONSIVE_SINGLE_CHOICE_ARRAY  = "responsive_single_choice_array"
QUESTION_TYPE_RESPONSIVE_BIPOLAR_LIKERT_ARRAY = "responsive_bipolar_likert_array"
QUESTION_TYPE_CLOZE                           = "cloze"
QUESTION_TYPE_UNKNOWN                         = "unknown"
QUESTION_TYPE_EMPTY                           = "empty"


QUESTION_TYPE_INPUTS = [QUESTION_TYPE_TEXT_INPUT, QUESTION_TYPE_DATE_INPUT, QUESTION_TYPE_NUMBER_INPUT]

OPTION_TYPE_DROPDOWN_OPTION             = "option"
OPTION_TYPE_RADIO                       = "radio"
OPTION_TYPE_CHECKBOX                    = "checkbox"
OPTION_TYPE_TEXT_INPUT                  = "text"
OPTION_TYPE_DATE_INPUT                  = "date"
OPTION_TYPE_NUMBER_INPUT                = "number"
OPTION_TYPE_CLOZE                       = "cloze"
OPTION_TYPE_DROPDOWN                    = "dropdown"
OPTION_TYPE_EMBEDDED_CLOZE_TEXT_INPUT   = "embedded_cloze_text"
OPTION_TYPE_EMBEDDED_CLOZE_DATE_INPUT   = "embedded_cloze_date"
OPTION_TYPE_EMBEDDED_CLOZE_NUMBER_INPUT = "embedded_cloze_number"
OPTION_TYPE_EMBEDDED_CLOZE_DROPDOWN     = "embedded_cloze_dropdown"

OPEN_FIELD_COL_SUFFIX = "open"

class SurveyPreviewOption:

    def __init__(self, key:str, option_type:str, label:Optional[str]):
        self.key = key
        self.option_type = option_type
        self.label = label
        self.element_index = None

    def is_embedded_cloze(self):
        return self.option_type.startswith(EMBEDDED_CLOZE_PREFIX)

    def option_value_type(self):
        opt_type = self.option_type
        if self.is_embedded_cloze():
            opt_type = opt_type.removeprefix(EMBEDDED_CLOZE_PREFIX)
        if opt_type == "date":
            return "date"
        if opt_type == "number":
            return "number"
        return 'text'

    def readable(self, o: List = None):
        if o is None:
            o = []
        if self.label is not None:
            label = self.label
        else:
            label = ""
        o.append("      - %s : %s %s" % (self.key, self.option_type, label))
    
class SurveyPreviewResponseGroup:
    def __init__(self, key:str, response_type:str):
        self.key = key
        self.response_type = response_type
        self.element_index = None
        self.options : List[SurveyPreviewOption] = []

    def add_option(self, option: SurveyPreviewOption):
        self.options.append(option)

    def readable(self, o: List = None):
        if o is None:
            o = []
        o.append("   - RespGroup '%s' (%s)" % (self.key, self.response_type))
        for option in self.options:
            option.readable(o)
   
class SurveyPreviewQuestion:
    
    def __init__(self, key:str, title:str, question_type:str):
        self.key = key
        self.element_index = None
        self.title = title
        self.question_type = question_type
        self.responses: List[SurveyPreviewResponseGroup] = []

    def add_response(self, response: SurveyPreviewResponseGroup):
        self.responses.append(response)

    def readable(self, o: List = None):
        if o is None:
            o = []
        o.append(" - Question '%s' (%s)" % (self.key, self.question_type))
        for r in self.responses:
            r.readable(o)
    
class SurveyPreview:
    def __init__(self, version:str, published:int, unpublished=None) -> None:
        self.version = version
        self.published = published
        self.unpublished = unpublished
        self.questions: List[SurveyPreviewQuestion] = []

    def add_question(self, question:SurveyPreviewQuestion):
        self.questions.append(question)
    
    def readable(self, o: List = None):
        if o is None:
            o = []
        o.append("Survey version '%s'" % self.version )
        for q in self.questions:
            q.readable(o)
        return o
     
def _extract(data:Dict, name:str, cls, alt=None):
    """
        Helper to extract value from a dictionnary with an optional alternate name
    
    """
    v = None
    if name in data:
        v = data[name]
    else:
        if alt is not None and alt in data:
            v = data[alt]
        else:
            raise ValueError("Expecting field : '%s'" % (name) )
    if not isinstance(v, cls):
        raise ValueError("Field '%s' must of class '%s'" % (name, cls)) 
    return v     

def preview_question_from_json(d:dict):
    key = _extract(d, 'key', str, 'id')
    question_type = _extract(d, 'questionType', str)
    title = _extract(d,'title', str)
    question = SurveyPreviewQuestion(key, title, question_type)
    rr = _extract(d, 'responses', list)
    for index, r in enumerate(rr):
        response = preview_response_from_json(r)
        response.element_index = index
        question.add_response(response)
    return question

def preview_response_from_json(d:dict):
    key = _extract(d, 'key', str, 'id') # Some fields have been renamed in new backend version
    response_type = _extract(d, 'responseTypes', str, alt="responseType")
    response = SurveyPreviewResponseGroup(key, response_type)
    if 'options' in d:
        rr = _extract(d, 'options', list)
        for index, r in enumerate(rr):
            option = preview_option_from_json(r)
            option.element_index = IndexError
            response.add_option(option)
    return response

def preview_option_from_json(d:dict):
    key = _extract(d, 'key', str, 'id')
    if 'label' in d:
        label = _extract(d, 'label', str)
    else:
        label = None
    option_type = _extract(d, 'optionType', str)
    option = SurveyPreviewOption(key, option_type, label)
    return option

def preview_from_json(d):
    """
        Build SurveyPreview model from json
        If dictionnary is provided, it's handled as a single survey preview
    
    """
    if isinstance(d, dict):
        version = _extract(d, 'versionId', str)
        unpublished = None
        published = None
        if 'published' in d:
            published = int(d['published'])
        if 'unpublished' in d:
            unpublished = int(d['unpublished'])
        survey = SurveyPreview(version, published, unpublished)
        qq = _extract(d, 'questions', list)
        for index, q in enumerate(qq):
            question = preview_question_from_json(q)
            question.element_index = index
            survey.add_question(question)
        return survey
    if isinstance(d, list):
        versions = []
        for item in d:
            version = preview_from_json(item)
            versions.append(version)
        return versions



from typing import Dict, Optional, List
from . import models as preview
from .models import SurveyPreviewOption, SurveyPreviewQuestion, SurveyPreviewResponseGroup, SurveyPreview

class SurveyRef:
    """
        SurveyRef keeps reference of the survey element providing a given column
    """
    def __init__(self, question, group_index=None, group_key=None, option=None, version=None):
        self.question = question
        self.group_index = group_index
        self.group_key = group_key
        self.option = option
        self.version = version

    def signature(self):
        """
        Return a function signature uniquely identifying reference in a survey
        The version is not part of a signature, since the same ref can exists in several surveys
        """
        s = self.question
        if self.group_key is not None:
            s += "|%s" % (self.group_key)

        if self.option is not None:
            s += "/" + self.option
        return s
    
    def to_dict(self):
        return {'question': self.question, 'group': [self.group_index, self.group_key], 'option': self.option, 'version': self.version}
    
class SurveyExportColumn:
    """
        SurveyExportColumn holds information about an exported column
    """

    def __init__(self, column_name:str, value_type:str, ref: SurveyRef):
        self.name = column_name
        self.value_type = value_type
        self.ref = ref
        self.seen_at = []

    def visit(self, version: str):
        self.seen_at.append(version)

    def to_dict(self):
        return {"name": self.name, "type": self.value_type, "ref": self.ref.to_dict(), "seen": self.seen_at}

class SchemaProblem:

    def __init__(self, ref:SurveyRef, context=None):
        self.ref = ref
        self.context = context

    def to_dict(self):
        return {"ref": self.ref.to_dict(), "context": self.context}

class SurveyExportSchema:
    """
        SurveyExportSchema holds the description of all exportable columns inferred from the survey preview
        A schema can be specific to a version, or be merge with several versions
        It also embeds list of eventual problems reported during the mapping process
    """
    def __init__(self, version=None):
        self.columns: Dict[str, SurveyExportColumn] = {}
        self.problems: Dict[str, List[SchemaProblem]] = {}
        self.version = version

    def register(self, column_name:str, value_type:str, ref: SurveyRef):
        if column_name in self.columns:
            prev = self.columns[column_name]
            problem_id = "duplicate:%s" % (column_name)
            self.report(problem_id, ref, {"prev":prev.ref})
            return
        self.columns[column_name] = SurveyExportColumn(column_name, value_type=value_type, ref=ref)

    def report(self, problem:str, ref: SurveyRef, context=None):
        if ref.version is None and self.version is not None:
            ref.version = self.version
        if problem not in self.problems:
            self.problems[problem] = []
        self.problems[problem].append(SchemaProblem(ref, context))

    def merge_problems(self, problems: Dict[str, List[SchemaProblem]]):
        for (id, pp) in problems.items():
            if id in self.problems:
                self.problems[id].extend(pp)
            else:
                self.problems[id] = pp

    def to_dict(self):
        cols = {}
        for name, column in self.columns.items():
            cols[name] = column.to_dict()
        problems = {}
        for name, probs in self.problems.items():
            pp = [ x.to_dict() for x in probs]
            problems[name] = pp
        return {"columns": cols, "problems": problems}
    
QUESTION_TYPES_SINGLE_GROUP = ['single_choice', 'multiple_choice']
OPTION_OPEN_TYPES = ['text']

class HandlerResponse:
    """
        Build list of known columns for a given response of a question
    """

    def __init__(self, separator: str, question: SurveyPreviewQuestion, single_slot: bool, value_type:Optional[str]=None):
        """
            value_type: optional type for the value (type of response)
            separator: question/option separator (given to response exporter)
            question: the question to handle
        """
        self.value_type = value_type
        self.separator = separator
        self.question = question
        single_rg = len(question.responses) == 1
        # If true, the response group key will be omited in the column name if there is only one response group (RG) for a question
        # It's the case for common question types, expecting usually only one RG, like single choice
        if single_slot and single_rg:
            self.use_rg_key = False
        else:
            self.use_rg_key = True
        
    def as_ref(self, rg: SurveyPreviewResponseGroup, option: Optional[SurveyPreviewOption]=None):
        """
            Creates a reference to the currently handled element of the survey
        """
        o_key = None
        if option is not None:
            o_key = option.key
        return SurveyRef(question=self.question.key, group_index=rg.element_index, group_key=rg.key, option=o_key)

    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        """
            Build response columns
        """
        pass

    def value_column(self, rg:SurveyPreviewResponseGroup):
        """
            Value column builds a name when a single value is expected for a question response
            The name of the column depends on the context if the question type handle the response group key or not
        """
        if self.use_rg_key:
            return self.question.key + self.separator + rg.key
        return self.question.key
        
    def option_column(self, rg:SurveyPreviewResponseGroup, id):
        """
             builds a column name for an option of a response group
        """
        if self.use_rg_key:
            prefix = self.question.key + self.separator + rg.key + '.'
        else:   
            prefix = self.question.key + self.separator
        return prefix + id


class HandlerSingleChoice(HandlerResponse):

    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        schema.register(self.value_column(rg), 'str', self.as_ref(rg))
        if len(rg.options) == 0:
            return
        for option in rg.options:
            if option.option_type not in ['radio', 'dropdown']:
                o_ref = self.as_ref(rg, option)
                # Other like input have extra columns
                # Option key can be composite (option key + '.' + sub cloze key )
                name = self.option_column(rg, option.key)
                schema.register(name, 'str', o_ref)

class HandlerValue(HandlerResponse):

    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        schema.register(self.value_column(rg), 'bool', self.as_ref(rg))

class HandlerMultipleChoice(HandlerResponse):
    
    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        for option in rg.options:
            o_ref = self.as_ref(rg, option)
            name = self.option_column(rg, option.key)
            is_embedded = option.is_embedded_cloze()
            column_type = 'bool'
            if is_embedded:
                column_type = option.option_value_type()
            schema.register(name, column_type, o_ref)
            # If it's not a checkbox, and not embedded, then it's an open option
            # create the associated column
            if option.option_type not in [preview.OPTION_TYPE_CHECKBOX] and not is_embedded:
                if option.option_type == preview.OPTION_TYPE_CLOZE:
                    open_name = self.option_column(rg, option.key)
                else:
                    open_name = name + self.separator + preview.OPEN_FIELD_COL_SUFFIX
                schema.register(open_name, 'str', o_ref)

class HandlerCloze(HandlerResponse):
    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        for option in rg.options:
            if option.option_type not in ['radio', 'dropdown']:
                o_ref = self.as_ref(rg, option)
                # Other like input have extra columns
                # Option key can be composite (option key + '.' + sub cloze key )
                name = self.option_column(rg, option.key)
                schema.register(name, 'str', o_ref)

class HandlerUnknown(HandlerResponse):
    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        schema.register(self.value_column(rg), 'json', self.as_ref(rg))

class HandlerEmpty(HandlerResponse):
    def build_response(self, rg: SurveyPreviewResponseGroup, schema: SurveyExportSchema):
        # Nothing to do
        pass

class HandlerQuestion:
    """
        Define a handler for a question
    """
    def __init__(self, handler:type[HandlerResponse], single_slot:bool, value_type:Optional[str]=None):
        self.handler = handler
        self.single_slot = single_slot
        self.value_type = value_type

    def build(self, question: SurveyPreviewQuestion, separator: str, schema: SurveyExportSchema):
        h = self.handler(separator=separator, question=question, single_slot=self.single_slot, value_type=self.value_type)
        for rg in question.responses:
            h.build_response(rg, schema) 

HANDLERS: Dict[str, HandlerQuestion] = {}

def handler_def(question_type:str, handler:type[HandlerResponse], single_slot:bool, value_type:Optional[str]=None):
    if question_type in HANDLERS:
        raise KeyError("Duplicate handler for %s" % question_type)
    HANDLERS[question_type] = HandlerQuestion(handler=handler, single_slot=single_slot, value_type=value_type)

def register_handlers():
    
    if len(HANDLERS) > 0:
        return

    # Single Choice and single choice like
    handler_def(preview.QUESTION_TYPE_SINGLE_CHOICE, HandlerSingleChoice, single_slot=True   ) 
    handler_def(preview.QUESTION_TYPE_DROPDOWN, HandlerSingleChoice, single_slot=True  )
    handler_def(preview.QUESTION_TYPE_LIKERT, HandlerSingleChoice, single_slot=True )
    handler_def(preview.QUESTION_TYPE_LIKERT_GROUP, HandlerSingleChoice, single_slot=True  )
    handler_def(preview.QUESTION_TYPE_RESPONSIVE_SINGLE_CHOICE_ARRAY, HandlerSingleChoice, single_slot=True  )
    handler_def(preview.QUESTION_TYPE_RESPONSIVE_BIPOLAR_LIKERT_ARRAY, HandlerSingleChoice, single_slot=True  )

    # Inputs
    handler_def(preview.QUESTION_TYPE_TEXT_INPUT, HandlerValue, single_slot=True, value_type='str')
    handler_def(preview.QUESTION_TYPE_NUMBER_INPUT, HandlerValue, single_slot=True, value_type='number')
    handler_def(preview.QUESTION_TYPE_NUMERIC_SLIDER, HandlerValue, single_slot=True, value_type='number')
    handler_def(preview.QUESTION_TYPE_DATE_INPUT, HandlerValue, single_slot=True, value_type='date')
    handler_def(preview.QUESTION_TYPE_EQ5D_SLIDER, HandlerValue, single_slot=True)
    handler_def(preview.QUESTION_TYPE_CONSENT, HandlerValue, single_slot=True, value_type='bool')
    
    # Multiple Choice
    handler_def(preview.QUESTION_TYPE_MULTIPLE_CHOICE, HandlerMultipleChoice, single_slot=True)
    
    # Responsive Table
    handler_def(preview.QUESTION_TYPE_RESPONSIVE_TABLE, HandlerValue, single_slot=False)
    
    # Matrix
    handler_def(preview.QUESTION_TYPE_MATRIX, HandlerValue, single_slot=False)
    handler_def(preview.QUESTION_TYPE_MATRIX_RADIO_ROW, HandlerValue, single_slot=False)
    handler_def(preview.QUESTION_TYPE_MATRIX_DROPDOWN, HandlerValue, single_slot=False)
    handler_def(preview.QUESTION_TYPE_MATRIX_INPUT, HandlerValue, single_slot=False)
    handler_def(preview.QUESTION_TYPE_MATRIX_NUMBER_INPUT, HandlerValue, single_slot=False, value_type='number')
    handler_def(preview.QUESTION_TYPE_MATRIX_CHECKBOX, HandlerValue, single_slot=False)

    # Cloze
    handler_def(preview.QUESTION_TYPE_CLOZE, HandlerMultipleChoice, single_slot=True)
    
    handler_def(preview.QUESTION_TYPE_UNKNOWN, HandlerUnknown, single_slot=False)
    
    handler_def(preview.QUESTION_TYPE_EMPTY, HandlerEmpty, single_slot=False)

    # case QUESTION_TYPE_SINGLE_CHOICE:
	# 	return processResponseForSingleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_CONSENT:
	# 	return processResponseForConsent(question, response, questionOptionSep)
	# case QUESTION_TYPE_DROPDOWN:
	# 	return processResponseForSingleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_LIKERT:
	# 	return processResponseForSingleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_LIKERT_GROUP:
	# 	return handleSingleChoiceGroupList(question.ID, question.Responses, response, questionOptionSep)
	# case QUESTION_TYPE_RESPONSIVE_SINGLE_CHOICE_ARRAY:
	# 	return processResponseForSingleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_RESPONSIVE_BIPOLAR_LIKERT_ARRAY:
	# 	return processResponseForSingleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_MULTIPLE_CHOICE:
	# 	return processResponseForMultipleChoice(question, response, questionOptionSep)
	# case QUESTION_TYPE_TEXT_INPUT:
	# 	return processResponseForInputs(question, response, questionOptionSep)
	# case QUESTION_TYPE_DATE_INPUT:
	# 	return processResponseForInputs(question, response, questionOptionSep)
	# case QUESTION_TYPE_NUMBER_INPUT:
	# 	return processResponseForInputs(question, response, questionOptionSep)
	# case QUESTION_TYPE_NUMERIC_SLIDER:
	# 	return processResponseForInputs(question, response, questionOptionSep)
	# case QUESTION_TYPE_EQ5D_SLIDER:
	# 	return processResponseForInputs(question, response, questionOptionSep)
	# case QUESTION_TYPE_RESPONSIVE_TABLE:
	# 	return processResponseForResponsiveTable(question, response, questionOptionSep)
	# case QUESTION_TYPE_MATRIX:
	# 	return processResponseForMatrix(question, response, questionOptionSep)
	# case QUESTION_TYPE_CLOZE:
	# 	return processResponseForCloze(question, response, questionOptionSep)
	# case QUESTION_TYPE_UNKNOWN:
	# 	return processResponseForUnknown(question, response, questionOptionSep)

class SurveySchemaBuilder:

    def __init__(self, separator:str, prefix: str):
        self.separator = separator
        self.prefix = prefix
        # Global schema intented to hold a merge of all versions
        self.schema = SurveyExportSchema()
        register_handlers()

    def build_survey(self, survey: SurveyPreview, merge_schema:bool=True):
        """
            Build an export schema from a survey preview
            if merge_schema is True, merge with global schema
        """
        version_schema = SurveyExportSchema(version=survey.version)
        for question in survey.questions:
            handler = HANDLERS.get(question.question_type, None)
            if handler is None:
                version_schema.report("unknown_handle:%s" % (question.question_type), SurveyRef(question=question.key))
                continue
            handler.build(separator=self.separator, question=question, schema=version_schema)
        if merge_schema:
            self.merge_schema(version_schema)
        return version_schema
        
    def merge_schema(self, schema: SurveyExportSchema):
        for (name, column) in schema.columns.items():
            if name in self.schema.columns:
                prev = self.schema.columns[name]
                if column.ref.signature() != prev.ref.signature():
                    self.schema.report("ref_changed:%s" % name, column.ref, {"message": "'%s' vs '%s'" % (column.ref.signature(), prev.ref.signature())})
                else:
                    prev.visit(schema.version)
            else:
                self.schema.columns[name] = column
        self.schema.merge_problems(schema.problems)

class ReadableSchema:
    """
        Build a readable represnetation of schema content as a dictionnary
        This aims at show a readable (with short content) view once serialized in json/yaml
        It's not intented to produce data to be stored, but mostly printed on console
    """
    def __init__(self):
        pass

    def build(self, schema: SurveyExportSchema):
        cols = {}
        for (name, column) in schema.columns.items():
            cols[name] = self.readable_column(column)
        problems = {}
        for (name, pp) in schema.problems.items():
            problems[name] = [ self.readable_problem(x) for x in pp ]
        return {"columns": cols, "problems": problems}
        
    def readable_column(self, column:SurveyExportColumn):
        return "[%s] from %s (%s)" % (column.value_type, self.readable_ref(column.ref), ','.join( column.seen_at))
    
    def readable_ref(self, ref: SurveyRef):
        if ref.version is not None:
            v = '@' + str(ref.version)
        else: 
            v = ''
        return ref.signature() + v
    
    def readable_problem(self, problem: SchemaProblem):
        return 'from ' + self.readable_ref(problem.ref)



from .context import Context

class RedeableRenderer:
    """
        Render a value (possibly an complex structure) to a human readable structure and render parsed elements
        
        This class handles the transformation of some elements parsed in object like Expression, Translatable
        The rendering here is just to transform some elements to a more friendly readable form once serialized in yaml
        It takes a context to be able to filter (for example only show one language).

        The intention here is not to maintain a machine readable structure, as some elements can be merged or output as string

    """

    def __init__(self, context:Context):
        self.context = context

    def render(self, value):
        if hasattr(value, 'to_readable'):
            r = value.to_readable(self.context)
            return self.render(r)

        if isinstance(value, dict):
            dd = {}
            for k, v in value.items():
                dd[k] = self.render(v)
            return dd
        if isinstance(value, list):
            dd = []
            for v in value:
                dd.append(self.render(v))
            return dd
        
        return value
    

def as_readable(value, context:Context):
    """
        Transform a structure to a readable one
        value should first be parsed using parsers function to prepare some elements into more useable objects
        see study_parser, parse_translatable, expression_parser
    """
    renderer = RedeableRenderer(context)
    return renderer.render(value)






from datetime import datetime

class Timestamp:

    def __init__(self, value):
        """
        docstring
        """
        self.value = int(value)

    def __repr__(self):
        return "<timestamp %d>" % self.value

    def to_time(self):
        return datetime.fromtimestamp(self.value)

    def to_readable(self, ctx):
        return self.to_time()


from typing import Dict, Union

from influenzanet.surveys.influenzanet.survey import Survey
from .influenzanet import survey_parser
from jinja2 import Template, FileSystemLoader, Environment
from .context import Context, create_context
from .templates.html import get_html_path
import os
import json
import glob
import traceback

known_styles = {
  'item': 'card',
  'translate-list': 'list-unstyled',
  'trans-code': 'badge badge-light',
  'components': 'card',
  'item-version': 'badge badge-success',
  'role': 'badge badge-warning',
}

def styles(name):
    if name in known_styles:
        return known_styles[name] + ' '+ name
    return name

def survey_to_html(survey:Union[Dict, Survey], context: Context):
    """
        Build an HTML document from a survey json
    """

    if not isinstance(survey, Survey):
        survey = survey_parser(survey)

    path = get_html_path()

    env = Environment(loader=FileSystemLoader(path), autoescape='html',)
    env.globals['language'] = context.get_language()
    env.globals['context'] = context
    env.globals['styles'] = styles
    
    template = env.get_template('survey.html')
    
    with open(path + '/survey.css') as f:
        theme = f.read()

    ctx = {
        'survey': survey,
        'theme_css': theme
    }
    return template.render(ctx)

def build_html_from_dir(path, languages):
    """"
        Helper function to build html files of all json files in a directory (recursively)
    """

    for f in glob.glob(path + "/*.json", recursive=True):
        print(f)
        survey = json.load(open(f, 'r', encoding='UTF-8'))

        if "studyKey" in survey:
            # A study entry
            survey = survey['survey']

        if not "props" in survey or not "surveyDefinition" in survey:
            print("%s is not a survey json" % f)
            continue
        try:
            h = survey_to_html(survey, context=create_context(language=languages))

            fn, ext = os.path.splitext(f)
            with open(fn + '.html', 'w') as o:
                o.write(h)
        except Exception as e:
            print("Unable to process '%s'" % (f))
            traceback.print_exception(e)





from collections import OrderedDict
import json
import yaml
from typing import Dict, List, Optional
from ..influenzanet.dictionnary import ItemDictionnary
from ..influenzanet.survey import SurveyItem, SurveySingleItem
from .. import standard
from ..standard.models import (MATRIX_CHOICE_TYPE, MULTIPLE_CHOICE_TYPE)
from ..standard.parser import json_parser_survey_standard
from ..utils import read_json

from .. import influenzanet
from ..standard import StandardQuestion, StandardSurvey

import requests

class ConfigError(Exception):
    pass

def read_yaml(path):
    obj = yaml.load(open(path, 'r', encoding='UTF-8'),  Loader=yaml.FullLoader)
    return obj

def read_from_url(url):
    try:
        response = requests.get(url)
        return response.content()
    except requests.RequestException as e:
        raise ConfigError("read_from_url(%s): %s" % (url, e, ))

class ValidatorProblem:

    MISSING = 'question_missing'
    NOT_DEFINED = 'question_unknown'
    WRONG_TYPE = 'wrong_type'
    MISSING_VALIDATION = 'missing_validation'
    OPT_NOT_DEFINED = 'option_unknown'
    OPT_MISSING = 'option_missing'

    def __init__(self, type:str, name:str, expected:Optional[str]=None, given:Optional[str]=None) -> None:
        self.type = type
        self.name = name
        self.given = given
        self.expected = expected
        self.known = False

    def to_readable(self, ctx):
        d = {'type': self.type, 'name': self.name, 'known': self.known}
        if self.expected is not None:
            d['expected'] = self.expected
        if self.given is not None:
            d['given'] = self.given
        return d

ValidatorProblem.TYPES = [
    ValidatorProblem.MISSING,
    ValidatorProblem.NOT_DEFINED,
    ValidatorProblem.WRONG_TYPE,
    ValidatorProblem.MISSING_VALIDATION,
    ValidatorProblem.OPT_MISSING,
    ValidatorProblem.OPT_NOT_DEFINED,
]

class ValidatorProfile:
    """
    ValidationProfile define parameters describing how to do the validation against the standard.
    It embeds parameters to map survey to the standard and options to mute some expected anomalies
    """
    def __init__(self):
        self.prefix = None
        self.standard = None
        self.standard_from = None
        self.expected = None

    @staticmethod
    def create(data):
        """
        Create the profile from a dictionnary (from a yaml or json file)

        Example:
        standard:
          # Full Path 
          file: 'path/to/standard.json' # Full name of the standard
          # Or directly from git repo
          name: 'standard_name' # name of the standard
          repo: 'influenzanet/surveys-standards' # repo name (optionnal, only if dont want default)
          revision: 'master' # revision in the repo, optionnal only if not master last commit
          # OR directly URL where to get the json standard

        prefix: 'key prefix to remove '
        expected: # Expected differences with the standard
          question_missing: [ "list of missing questions or pattern like Q10c*..."]
          question_unknown: [ "list of not expected questions or pattern like Q10c*..."]

        """
        p = ValidatorProfile()
        
        prefixes = data['prefix']
        if isinstance(prefixes, str):
            prefixes = [prefixes]
        p.prefixes = prefixes

        if not 'standard' in data:
            raise ConfigError("standard entry expected")
        std = data['standard']
        p.standard_from = std
        if 'expected' in data:
            p.expected = p.load_expected(data['expected'])
        return p

    def load_expected(self, expected:dict)-> Dict:
        for n, v in expected.items():
            if not n in ValidatorProblem.TYPES:
                raise ConfigError("Unknown problem %s in expected" % (n, ))
        return expected            

    def load_standard(self):
        std = self.standard_from
        standard_json = None
        url = None
        if 'file' in std:
            try:
                standard_json = read_json(std['file'])
            except Exception as e:
                raise ConfigError("Unable to load standard from '%s' %s" % (std['file'], e) )
        if 'name' in std:
            name = std['name']
            if 'repo' in std:
                repo = std['repo']
            else:
                repo = 'influenzanet/surveys-standards'
            if 'revision' in std:
                revision = std['revision']
            else:
                revision = 'master'
            url = 'https://github.com/%s/blob/%s/surveys/%s/survey.json?raw=true' % (repo, revision, name)
        if 'url' in std:
            url = std['url']
        if url is not None:
            standard_json = json.loads(read_from_url(url))
            #print(standard_json)

        self.standard = json_parser_survey_standard(standard_json)

COMPATIBLE_TYPES = {
    standard.SINGLE_CHOICE_TYPE : [influenzanet.RGROLES.SINGLE, influenzanet.RGROLES.DROPDOWN],
    standard.MULTIPLE_CHOICE_TYPE : influenzanet.RGROLES.MULTIPLE,
    standard.DATE_TYPE: influenzanet.RGROLES.DATE,
    standard.MATRIX_CHOICE_TYPE: influenzanet.RGROLES.MATRIX,
}

class SurveyStandardValidator:
    """"Validate a survey definition compliance to a standard survey description

    """

    def __init__(self, profile: ValidatorProfile):
        """
            profile: ValidatorProfile
            Structure embedding information about how to process the validation
            Which standard and where to find it, and validation options
        
        """
        self.profile = profile
        self.standard = profile.standard

        
    @staticmethod
    def profile_from_yaml(file: str):
        """
            Read Profile from yaml file
            Validation profile is expected to be under a 'profile' entry
        """
        p = read_yaml(file)
        profile = ValidatorProfile.create(p['profile'])
        profile.load_standard()
        return profile
    
    def validate(self, definition:influenzanet.SurveyItem):
        """Validate survey definition to the standard

        Parameters
        ------
            definition: influenzanet.SurveyItem
                Survey definition to validate (usually the "current" component)

            options: ValidatorProfile
                Validation profile describes how to do the validation
        """

        problems = []
        
        expected = OrderedDict()
        for quid, q in self.standard.questions.items():
            if q.active:
                expected[quid] = q
        items = definition.flatten()
        for item in items:
            if item.is_group():
                continue
            rg = item.get_dictionnary()
            if rg is None:
                # Not a question
                continue

            item_key = item.key
            
            item_key = self.remove_prefix(item_key)
            if not item_key in expected:
                problems.append(ValidatorProblem(ValidatorProblem.NOT_DEFINED, item_key))
                continue
            std = expected[item_key]
            self.compare_question(item, rg, std, problems)
            del expected[item_key]
            
        if len(expected) > 0:
            # Question not found, mark them as missing
            for e in expected:
                problems.append(ValidatorProblem(ValidatorProblem.MISSING, e))

        # Flag problems with expected rules
        self.validate_problems(problems)

        return problems

    def remove_prefix(self,  item_key):
        """
            Remove prefix for item key from registred ones
        """
        if self.profile.prefixes is None:
            return item_key
        for prefix in self.profile.prefixes:
            if item_key.startswith(prefix):
                return item_key[len(prefix):]
        return item_key

    def validate_problems(self, problems:List[ValidatorProblem]):
        """
            Flag problems as known considering expected rules
            rules can be a question, a pattern (* for starting with), or a option with the form [question].[option key]
        """
        expected = self.profile.expected
        if expected is None:
            return
        for problem in problems:
            for pb_type in ValidatorProblem.TYPES:
                if(problem.type == pb_type):
                    if self.problem_in_list(expected, pb_type , problem.name):
                        problem.known = True

    def problem_in_list(self, expected: Dict, problem_type, name):
        knowns = expected.get(problem_type, None)
        if knowns is None:
            return False
        for known in knowns:
            if known.find('*') >= 0:
                pattern = known.replace('*', '')
                return name.startswith(pattern)
            else:
                if name == known:
                    return True
        return False

    def compare_question(self, item:SurveySingleItem, rg: Optional[ItemDictionnary], std: StandardQuestion, problems: List[ValidatorProblem]):
        expected_types = COMPATIBLE_TYPES.get(std.type, None)
        if expected_types is not None:
            if isinstance(expected_types, list):
                found = rg.type in expected_types
            else:
                found = rg.type == expected_types
            if not found:
                problems.append(ValidatorProblem(ValidatorProblem.WRONG_TYPE, std.data_name, expected_types, rg.type))
        if std.mandatory:
            if item.validations is None:
                problems.append(ValidatorProblem(ValidatorProblem.MISSING_VALIDATION, item.key))
        
        if len(std.responses) == 0:
            return
        expected_keys = OrderedDict()
        for r in std.responses.values():
            expected_keys[r.value] = r
        if rg.options is not None:
            for r in rg.options:
                if r.item_key in expected_keys:
                    del expected_keys[r.item_key]
                else:
                    problems.append(ValidatorProblem(ValidatorProblem.OPT_NOT_DEFINED, std.data_name + '.' + r.item_key))
        
        for name,k in expected_keys.items():
            problems.append(ValidatorProblem(ValidatorProblem.OPT_MISSING, std.data_name + '.' + name))
                
    def filter_known(self, problems: List[ValidatorProblem])->List[ValidatorProblem]:
        pp = []
        for p in problems:
            if p.known:
                continue
            pp.append(p)
        return pp




class RGRole:
    """
    Roles of ResponseGroup
    """
    def __init__(self, name, data:bool) -> None:
        self.name = name
        self.data = data

    def __eq__(self, value):
        if isinstance(value, str):
            return value == self.name
        if isinstance(value, self.__class__):
            return self.name == value.name
        return False

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return '<%s,%s>' % (self.__class__.name, self.name)

    def to_readable(self, ctx):
        return self.name

class RGROLES:
    # Display Roles
    TEXT = RGRole('text', False)
    MARKDOWN = RGRole('markdown', False)
    
    # Data Input Roles
    SINGLE = RGRole('singleChoiceGroup', True)
    DROPDOWN = RGRole('dropDownGroup', True)
    MATRIX = RGRole('matrix', True)
    MULTIPLE = RGRole('multipleChoiceGroup', True)
    DATE = RGRole('dateInput', True)
    INPUT = RGRole('input', True)
    MULTIPLELINE=RGRole('multilineTextInput', True)
    NUMBER = RGRole('numberInput', True)
    SLIDER_NUM = RGRole('sliderNumeric', True)
    SLIDER_RANGE = RGRole('sliderNumericRange', True)
    SLIDER_CAT = RGRole('sliderCategorical', True)
    LIKERT = RGRole('likert', True)
    LIKERTGROUP = RGRole('likertGroup', True)
    BIPOLARLIKERT = RGRole('responsiveBipolarLikertScaleArray', True)
    SINGLE_ARRAY = RGRole('responsiveSingleChoiceArray', True)

"""
    Known ResponseGroup Roles
"""
RG_ROLES = [v for k,v in RGROLES.__dict__.items() if isinstance(v,RGRole)]

"""
    Roles for Response Groups holding Data input
"""
RG_ROLES_DATA = [v for v in RG_ROLES if v.data]

"""
    Dictionnary associating role name (provided in json) and internal class
"""
RG_ROLES_DICT = zip( [v.name, v ] for v in RG_ROLES)



from typing import List, Optional

class Context:
    """
    Rendering context. This class embeds parameters used in renderding and transforming layers
    For example the languages to show in the output. 
    """
    def __init__(self, language:Optional[List[str]]=None):
        self.language = language
    
    def get_language(self):
        return self.language

def create_context(language=None):
    if language is not None:
        if isinstance(language, str):
            language = language.split(',')
        if not isinstance(language, list):
            raise Exception("language must be a list (or a string comma separated items)")
        language = [x.strip() for x in language]
    return Context(language=language)



import json

def read_json(path):
    data = json.load(open(path, 'r', encoding='UTF-8'))
    return data

def translatable_to_list(data, language=None):
    values = []
    for d in data:
        s = []
        if language is not None:
            if d['code'] != language:
                continue
        else:
            s.append("[%s] " % d['code'])
        for p in d['parts']:
            s.append(p['str'])
        values.append(' '.join(s))
    return values        
